"""add channel size limit

Revision ID: db1c56bf4d57
Revises: a3ffa287d074
Create Date: 2020-12-08 18:44:14.215836

"""
import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = 'db1c56bf4d57'
down_revision = 'a3ffa287d074'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column('channels', sa.Column('size', sa.Integer(), nullable=True))
    op.add_column('channels', sa.Column('size_limit', sa.Integer(), nullable=True))
    op.add_column('package_versions', sa.Column('size', sa.Integer(), nullable=True))
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column('package_versions', 'size')
    op.drop_column('channels', 'size_limit')
    op.drop_column('channels', 'size')
    # ### end Alembic commands ###
